﻿using System;
using System.Diagnostics;
using System.Threading;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Widgets;
using OpcLabs.UnitTestLib;

namespace OpcLabs.BaseLibTest.Widgets
{
    public partial class NotifyingWidgetTest
    {
        private class TestNotifyingWidget : NotifyingWidget
        {
            public void RaiseTestEvent([NotNull] EventArgs eventArgs)
            {
                EventHandler<EventArgs> handler = TestEvent;
                Raise((handler == null) ? null : new Handler<EventArgs>(handler), eventArgs);
            }

            public EventHandler<EventArgs> TestEvent;
        }



        #region Test Methods

        [TestMethod]
        public void RaiseTest()
        {
            TestSection.Arrange();

            var target = new TestNotifyingWidget();
            Debug.Assert(!target.QueueCallbacks);

            bool eventRaised = false;
            object raisedSender = null;
            EventArgs raisedEventArgs = null;
            target.TestEvent += (sender, args) =>
            {
                raisedSender = sender;
                raisedEventArgs = args;
                eventRaised = true;
            };

            TestSection.Act();
            var testEventArgs = new EventArgs();
            target.RaiseTestEvent(testEventArgs);

            TestSection.Assert();
            Assert.IsTrue(eventRaised);
            Assert.AreSame(target, raisedSender);
            Assert.AreSame(testEventArgs, raisedEventArgs);
        }

        [TestCategory("TimeDomain")]
        [TestMethod]
        public void RaiseTest_QueueCallbacks_OneEvent()
        {
            TestSection.Arrange();

            var target = new TestNotifyingWidget {QueueCallbacks = true};

            ManualResetEvent eventRaisedEvent = new ManualResetEvent(initialState:false);
            object raisedSender = null;
            EventArgs raisedEventArgs = null;
            target.TestEvent += (sender, args) =>
            {
                raisedSender = sender;
                raisedEventArgs = args;
                eventRaisedEvent.Set();
            };

            TestSection.Act();
            var testEventArgs = new EventArgs();
            target.RaiseTestEvent(testEventArgs);
            bool eventRaised = eventRaisedEvent.WaitOne(2 * 1000);

            TestSection.Assert();
            Assert.IsTrue(eventRaised);
            Assert.AreSame(target, raisedSender);
            Assert.AreSame(testEventArgs, raisedEventArgs);
        }

        [TestCategory("TimeDomain")]
        [TestMethod]
        public void RaiseTest_QueueCallbacks_TwoEvents()
        {
            RaiseTest_QueueCallbacks(
                callbackQueueIdleTimeToSleep: NotifyingWidget.DefaultCallbackQueueIdleTimeToSleep,
                delayBetweenEvents: 0);
        }

        [TestCategory("TimeDomain")]
        [TestMethod]
        public void RaiseTest_QueueCallbacks_TwoEvents_AfterIdleTimeToSleep()
        {
            RaiseTest_QueueCallbacks(
                callbackQueueIdleTimeToSleep: 4*1000,
                delayBetweenEvents: 10*1000);
        }

        #endregion



        #region Private Methods

        public void RaiseTest_QueueCallbacks(int callbackQueueIdleTimeToSleep, int delayBetweenEvents)
        {
            TestSection.Arrange();

            var target = new TestNotifyingWidget { QueueCallbacks = true, CallbackQueueIdleTimeToSleep = callbackQueueIdleTimeToSleep };

            ManualResetEvent eventRaisedEvent = new ManualResetEvent(initialState: false);
            object raisedSender = null;
            EventArgs raisedEventArgs = null;
            target.TestEvent += (sender, args) =>
            {
                raisedSender = sender;
                raisedEventArgs = args;
                eventRaisedEvent.Set();
            };

            TestSection.Act();

            var testEventArgs1 = new EventArgs();
            target.RaiseTestEvent(testEventArgs1);
            bool eventRaised1 = eventRaisedEvent.WaitOne(2 * 1000);
            object raisedSender1 = raisedSender;
            EventArgs raisedEventArgs1 = raisedEventArgs;

            AssertUtilities.WaitingPhase(delayBetweenEvents);

            eventRaisedEvent.Reset();
            var testEventArgs2 = new EventArgs();
            target.RaiseTestEvent(testEventArgs2);
            bool eventRaised2 = eventRaisedEvent.WaitOne(2 * 2000);
            object raisedSender2 = raisedSender;
            EventArgs raisedEventArgs2 = raisedEventArgs;

            TestSection.Assert();

            Assert.IsTrue(eventRaised1, "First event not raised");
            Assert.AreSame(target, raisedSender1);
            Assert.AreSame(testEventArgs1, raisedEventArgs1);

            Assert.IsTrue(eventRaised2, "Second event not raised");
            Assert.AreSame(target, raisedSender2);
            Assert.AreSame(testEventArgs2, raisedEventArgs2);
        }

        #endregion
    }
}
