﻿
using System;
using System.Collections.Generic;
using System.Threading;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.ComInterop;
using OpcLabs.BaseLib.Runtime;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.OperationModel;
using OpcLabs.EasyOpc.Engine;
using OpcLabs.EasyOpc.Implementation.NetApi;
using OpcLabs.EasyOpcLiveTest.DataAccess.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.DataAccess
{
    partial class EasyDAClientTestBase
    {
        protected void SubscribeItemCatalogTest_Basic(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] DAItemCatalogEntry itemCatalogEntry)
        {
            VarType dataType = itemCatalogEntry.DataType;
            Type expectedType = null;
            if (dataType != VarTypes.Empty)
                expectedType = Cls.GetClsType(VarTypeUtilities.TypeFromVarType(dataType));

            SubscribeItemTest_Basic(serverDescriptor, itemCatalogEntry.Element, expectedType);
        }

        protected void SubscribeItemTest_Error(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] DAItemDescriptor itemDescriptor)
        {
            TestSection.Arrange();
            AssertUtilities.WriteLine(serverDescriptor);
            AssertUtilities.WriteLine(itemDescriptor);
            AssertUtilities.WriteLine("");

            TestEnvironment.PrecheckServer(serverDescriptor);
            var client = new EasyDAClient();

            var senderList = new List<object>();
            var argsList = new List<EasyDAItemChangedEventArgs>();
            var completeEvent = new AutoResetEvent(initialState: false);

            void ClientOnItemChanged(object sender, EasyDAItemChangedEventArgs args)
            {
                AssertUtilities.WriteLine(args);
                senderList.Add(sender);
                argsList.Add(args);

                completeEvent.Set();
            }

            const int requestedUpdateRate = 100;
            const int testState = 777777777;

            TestSection.Act();
            client.ItemChanged += ClientOnItemChanged;
            int handle = client.SubscribeItem(serverDescriptor, itemDescriptor, requestedUpdateRate, testState);
            bool completed = completeEvent.WaitOne(5 * 1000);
            client.UnsubscribeItem(handle);
            client.ItemChanged -= ClientOnItemChanged;

            TestSection.Assert();

            AssertUtilities.WriteLine(argsList.Count);
            Assert.IsTrue(completed, "Not completed.");
            Assert.AreEqual(1, argsList.Count);

            {
                object sender = senderList[0];
                Assert.IsNotNull(sender);
                Assert.AreSame(client, sender);

                EasyDAItemChangedEventArgs args = argsList[0];
                Assert.IsNotNull(args);

                Assert.AreEqual(testState, args.Arguments.State); // is correct for value types only
                Assert.IsNull(args.Arguments.Callback);

                Assert.AreEqual(new DAGroupParameters(requestedUpdateRate), args.Arguments.GroupParameters);
                Assert.AreEqual(itemDescriptor, args.Arguments.ItemDescriptor);
                Assert.AreEqual(serverDescriptor, args.Arguments.ServerDescriptor);

                Exception exception = args.Exception;
                Assert.IsNotNull(exception);
                AssertUtilities.WriteLine(exception.Message ?? String.Empty);
                AssertUtilities.WriteLine(args.ErrorCode);
                Assert.IsTrue(args.ErrorCode < 0);
                Assert.IsNull(args.Vtq);

                Assert.IsInstanceOfType(exception, typeof(NetApiResultException));
                StringAssert.Contains(exception.Message, "\"E_UNKNOWN_ITEM_NAME\"");
            }
        }



        private void SubscribeItemTest_Basic(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] DAItemDescriptor itemDescriptor,
            [CanBeNull] Type expectedType)
        {
            TestSection.Arrange();
            AssertUtilities.WriteLine(serverDescriptor);
            AssertUtilities.WriteLine(itemDescriptor);
            AssertUtilities.WriteLine(expectedType);
            AssertUtilities.WriteLine("");

            TestEnvironment.PrecheckServer(serverDescriptor);

            var senderList = new List<object>();
            var argsList = new List<EasyDAItemChangedEventArgs>();
            int indexSet = -1;
            var completeEvent = new AutoResetEvent(initialState: false);

            void ClientOnItemChanged(object sender, EasyDAItemChangedEventArgs args)
            {
                AssertUtilities.WriteLine(args);
                senderList.Add(sender);
                argsList.Add(args);

                if (args != null)
                {
                    if ((args.Exception != null) || ((args.Vtq != null) && !args.Vtq.Quality.IsBad))
                    {
                        completeEvent.Set();
                        indexSet = argsList.Count - 1;
                    }
                }
            }

            const int requestedUpdateRate = 100;
            const int testState = 777777777;

            var client = new EasyDAClient();

            TestSection.Act();
            client.ItemChanged += ClientOnItemChanged;
            int handle = client.SubscribeItem(serverDescriptor, itemDescriptor, requestedUpdateRate, testState);
            bool completed = completeEvent.WaitOne(5 * 1000);
            client.UnsubscribeItem(handle);
            client.ItemChanged -= ClientOnItemChanged;

            TestSection.Assert();

            AssertUtilities.WriteLine(argsList.Count);
            Assert.IsTrue(completed, "Not completed.");

            if (argsList.Count == 1)
            {
                EasyDAItemChangedEventArgs args = argsList[0];
                // ReSharper disable once MergeSequentialChecks
                if ((args != null) && (args.Exception is NotImplementedException))
                    return; // test success
            }

            for (int i = 0; i < argsList.Count; i++)
            {
                object sender = senderList[i];
                Assert.IsNotNull(sender);
                Assert.AreSame(client, sender);

                EasyDAItemChangedEventArgs args = argsList[i];
                Assert.IsNotNull(args);

                Assert.AreEqual(testState, args.Arguments.State); // is correct for value types only
                Assert.IsNull(args.Arguments.Callback);

                Assert.AreEqual(new DAGroupParameters(requestedUpdateRate), args.Arguments.GroupParameters);
                Assert.AreEqual(itemDescriptor, args.Arguments.ItemDescriptor);
                Assert.AreEqual(serverDescriptor, args.Arguments.ServerDescriptor);

                Assert.IsNull(args.Exception);
                Assert.IsTrue(args.ErrorCode >= 0);

                DAVtq actualVtq = args.Vtq;
                Assert.IsNotNull(actualVtq);
                AssertUtilities.WriteLine(actualVtq);

                object actualValue = actualVtq.Value;

                if ((indexSet == -1) || (i < indexSet))
                {
                    Assert.IsTrue(actualVtq.Quality.IsBad);
                    Assert.IsNull(actualValue);
                }
                else
                {
                    Assert.IsTrue(actualVtq.Quality.IsGood);
                    if (expectedType != null)
                    {
                        Assert.IsNotNull(actualValue);
                        Assert.AreEqual(expectedType, actualValue.GetType());
                    }
                }

                Assert.AreEqual(DateTimeKind.Utc, actualVtq.Timestamp.Kind);
                Assert.AreNotEqual(actualVtq.Timestamp, DateTime.MinValue);
                Assert.AreNotEqual(actualVtq.Timestamp, DateTime.MaxValue);
                Assert.IsTrue(actualVtq.Timestamp >= DateTime.FromFileTimeUtc(0));
                DateTime utcNow = DateTime.UtcNow;
                AssertUtilities.WriteLine($"UtcNow: {utcNow}");
                TimeSpan timeTolerance = TimeSpan.FromSeconds(60);
                Assert.IsTrue(actualVtq.Timestamp <= utcNow + timeTolerance, "actualVtq.Timestamp <= utcNow + timeTolerance");
            }
        }
    }
}
