﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Runtime.InteropServices;
using System.Threading;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib;
using OpcLabs.BaseLib.Execution;
using OpcLabs.BaseLib.Net.Sockets;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.AddressSpace;
using OpcLabs.EasyOpc.DataAccess.OperationModel;
using OpcLabs.EasyOpc.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.DataAccess
{
    [TestClass]

    public partial class EasyDAClientTest
    {
        protected class WcfClientStaticEnvironment : StaticEnvironment
        {
            public WcfClientStaticEnvironment()
            {
                EasyDAClient.EnableNativeClient = false;
                EasyDAClient.EnableNetApiClient = false;
                EasyDAClient.EnableServiceClient = true;
            }
        }



        #region Static Public Methods

        static public void SubscribeMultipleItems_Incrementing([NotNull] string machineName, int itemCount)
        {
            TestSection.Arrange();

            ServerDescriptor serverDescriptor = new ServerDescriptor(machineName, "OPCLabs.KitServer.2");
            UA.LegacyUATestEnvironment.Current.PrecheckEndpoint(serverDescriptor);

            var eventArgsList = new List<EasyDAItemChangedEventArgs>();
            bool acceptingNotifications = true;

            var argumentsArray = new EasyDAItemSubscriptionArguments[itemCount];
            const int phaseCount = 100;
            for (int i = 0; i < itemCount; i++)
            {
                int copy = i / phaseCount;
                int phase = i % phaseCount;
                string itemId = String.Format(CultureInfo.InvariantCulture,
                    "Simulation.Incrementing (10 s).Copy_{0}.Phase_{1}", copy + 1, phase + 1);

                argumentsArray[i] = new EasyDAItemSubscriptionArguments(
                    // ReSharper disable RedundantArgumentName
                    serverDescriptor: serverDescriptor,
                    itemDescriptor: itemId,
                    groupParameters: 60 * 1000,
                    state: i,
                    // ReSharper disable once AccessToModifiedClosure
                    callback: (sender, args) =>
                    {
                        if (acceptingNotifications) eventArgsList.Add(args);
                    });
                // ReSharper restore RedundantArgumentName
            }

            IEasyDAClient target = new EasyDAClient();

            TestSection.Act();

            AssertUtilities.Phase("Subscribing");
            var stopwatch = new Stopwatch();
            stopwatch.Start();
            target.SubscribeMultipleItems(argumentsArray); // should not crash here
            stopwatch.Stop();
            AssertUtilities.WriteLine($"Elapsed: {stopwatch.Elapsed}");

            // Give time for notifications to arrive
            //AssertUtilities.WaitFor(5*1000 + (int)(itemCount*0.01)*1000);
            AssertUtilities.WaitingPhase(80 * 1000);

            AssertUtilities.Phase("Unsubscribing");
            target.UnsubscribeAllItems();

            acceptingNotifications = false;
            AssertUtilities.WaitingPhase(2 * 1000);

            TestSection.Assert();

            int count = eventArgsList.Count;
            AssertUtilities.WriteLine($"Count: {count}");
            Assert.AreNotEqual(0, count);

            foreach (EasyDAItemChangedEventArgs eventArgs in eventArgsList)
                AssertUtilities.WriteLine(eventArgs);
            AssertUtilities.WriteLine(".");

            AssertUtilities.BeginWriteLine();
            Trace.Write("Checking... ");
            for (int i = 0; i < count; i++)
            {
                EasyDAItemChangedEventArgs eventArgs = eventArgsList[i];
                Trace.Write(i + " ");

                Assert.IsNotNull(eventArgs);
                if (eventArgs.Exception != null)
                {
                    AssertUtilities.WriteLine(eventArgs.Arguments.ItemDescriptor);
                    AssertUtilities.WriteLine(eventArgs.Exception);
                }
                Assert.IsNull(eventArgs.Exception);
                // IMPROVE: test more
            }
            AssertUtilities.EndWriteLine();
            Assert.IsTrue(count >= itemCount);

            // IMPROVE: test what really happens and how it works internally...
        }

        static public void SubscribeMultipleItems_UnknownItems(int itemCount)
        {
            const string serverClass = "OPCLabs.KitServer.2";
            const int unknownItemId =
                unchecked((int)0xC0040007 /*The item is no longer available in the server address space.*/);

            var eventArgsList = new List<EasyDAItemChangedEventArgs>();

            int lastNotificationTick = -1;
            var argumentsArray = new EasyDAItemSubscriptionArguments[itemCount];
            for (int i = 0; i < itemCount; i++)
            {
                argumentsArray[i] = new EasyDAItemSubscriptionArguments(
                    // ReSharper disable RedundantArgumentName
                    machineName: "",
                    serverClass: serverClass,
                    itemId: String.Format(CultureInfo.InvariantCulture, "BFLMPSVZ.UnknownItem_{0}", i),
                    requestedUpdateRate: 10 * 1000,
                    state: i,
                    callback: (sender, args) =>
                    {
                        lastNotificationTick = System.Environment.TickCount;
                        eventArgsList.Add(args);
                    });
            }
            // ReSharper restore RedundantArgumentName

            IEasyDAClient target = new EasyDAClient();

            AssertUtilities.Phase("Subscribing");
            var stopwatch = new Stopwatch();
            stopwatch.Start();
            target.SubscribeMultipleItems(argumentsArray); // should not crash here
            stopwatch.Stop();
            AssertUtilities.WriteLine($"Tick count: {System.Environment.TickCount}");
            AssertUtilities.WriteLine($"Elapsed: {stopwatch.Elapsed}");

            // Give time for notifications to arrive
            AssertUtilities.WaitingPhase(20 * 1000);

            AssertUtilities.Phase("Unsubscribing");
            target.UnsubscribeAllItems();
            AssertUtilities.WriteLine($"Tick count: {System.Environment.TickCount}");

            AssertUtilities.WaitingPhase(5 * 1000);
            AssertUtilities.WriteLine($"Tick count: {System.Environment.TickCount}");

            int notificationCount = eventArgsList.Count;
            AssertUtilities.WriteLine($"Notification count: {notificationCount}");
            AssertUtilities.WriteLine($"Last notification tick: {lastNotificationTick}");
            Assert.AreNotEqual(0, notificationCount);

            foreach (EasyDAItemChangedEventArgs eventArgs in eventArgsList)
                AssertUtilities.WriteLine(eventArgs);
            AssertUtilities.WriteLine(".");

            AssertUtilities.BeginWriteLine();
            Trace.Write("Checking... ");
            for (int i = 0; i < notificationCount; i++)
            {
                EasyDAItemChangedEventArgs eventArgs = eventArgsList[i];
                Trace.Write(i + " ");

                Assert.IsNotNull(eventArgs);
                Exception exception = eventArgs.Exception;
                Assert.IsNotNull(exception);

                Assert.AreEqual(unknownItemId, exception.HResult);
                Assert.AreEqual(serverClass, exception.Source);

                if (exception is COMException comException)
                {
                    Assert.AreEqual(unknownItemId, comException.ErrorCode);
                    continue;
                }

                if (exception is NormalizedException)
                {
                    continue;
                }

                Assert.Fail($"Incorrect exception type: {exception.GetType()}");
            }
            AssertUtilities.EndWriteLine();
            Assert.AreEqual(itemCount, notificationCount);

            // IMPROVE: test what really happens and how it works internally...
        }

        #endregion



        #region Static Private Methods

        static private void WcfClientPrecheck()
        {
            EndpointDescriptor endpointDescriptor = EasyDAClient.ServiceProviderEndpointDescriptor;
            if (!TcpClientUtilities.CanCreate(endpointDescriptor))
                Assert.Inconclusive(
                    $"Failed pre-check: cannot create TCP client to service provider endpoint \"{endpointDescriptor}\".");
        }

        #endregion



        [Timeout(60 * 1000)]
        [TestCategory("Live")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void PullMultipleItemChangesTest_SubscribeItem()
        {
            var client = new EasyDAClient {PullItemChangedQueueCapacity = 1000};

            int handle = client.SubscribeItem("", "OPCLabs.KitServer.2", "Demo.Ramp", 1000);

            int count = 0;
            int startTickCount = System.Environment.TickCount;
            do
            {
                EasyDAItemChangedEventArgs[] eventArgsArray =
                    client.PullMultipleItemChanges(maximumCount:int.MaxValue, millisecondsTimeout: 2 * 1000);
                if (eventArgsArray == null)
                    continue;
                foreach (EasyDAItemChangedEventArgs eventArgs in eventArgsArray)
                {
                    AssertUtilities.WriteLine(eventArgs);
                    count++;

                    Assert.IsNotNull(eventArgs);
                    Exception exception = eventArgs.Exception;
                    Assert.IsNull(exception);
                    // IMPROVE: assert more about the event
                }
            } while (System.Environment.TickCount < startTickCount + 10 * 1000);

            client.UnsubscribeItem(handle);

            AssertUtilities.WriteLine(count);
            Assert.IsTrue(count >= 1);
        }

        [ExpectedException(typeof(InvalidOperationException))]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void PullMultipleItemChangesTest_ZeroCapacity()
        {
            var client = new EasyDAClient();
            client.PullMultipleItemChanges(maximumCount: int.MaxValue, millisecondsTimeout: 2000);
        }



        #region Test Methods

        [Ignore] // IMPROVE: currently, can only run as a solo test
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ClientParametersTest_ExecutionContext_InprocServer()
        {
            InternalClientParametersTest_ExecutionContext_InprocServer();
        }

        // this does not work yet, because static params cannot be transferred to web service.
        // IMPROVE: make the ExecutionContext part of the ServerDescriptor, or at least instance/adaptable params
        //[TestCategory("Live")]
        //[TestMethod]
        //public void ClientParametersTest_ExecutionContext_InprocServer_WcfClient()
        //{
        //    WcfClientPrecheck();
        //    using (new WcfClientStaticEnvironment())
        //        InternalClientParametersTest_ExecutionContext_InprocServer();
        //}

        [Ignore]    // IMPROVE: put back after making clear how the analyzer should be configured for this
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseLeavesTest_OpcAnalyzer_ProgId()
        {
            var target = new EasyDAClient();

            DANodeElementCollection actualResultCollection = target.BrowseNodes(
                new ServerDescriptor("OpcTestLab.OpcAnalyzer.1"),
                "",
                new DABrowseParameters(DABrowseFilter.All));
            Assert.IsNotNull(actualResultCollection);
            // IMPROVE: assert more
        }

        [Ignore] // IMPROVE: fails
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseNodesTest_BrowseNodesChunk_5()
        {
            BrowseNodesTest_BrowseNodesChunk(5);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_UnknownResource()
        {
            InternalReadItemTest_UnknownResource();
        }

        [Ignore] // IMPROVE: put back 
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_UnknownResource_WcfClient()
        {
            WcfClientPrecheck();
            using (new WcfClientStaticEnvironment())
                InternalReadItemTest_UnknownResource();
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_UnregisteredProgID()
        {
            InternalReadItemTest_UnregisteredProgID();
        }

        [Ignore] // IMPROVE: put back 
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_UnregisteredProgID_WcfClient()
        {
            WcfClientPrecheck();
            using (new WcfClientStaticEnvironment())
                InternalReadItemTest_UnregisteredProgID();
        }

        // Used for manually testing the internal behavior - whether OPC Groups are properly managed
        [Ignore] // IMPROVE: fails
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        // ReSharper disable InconsistentNaming
        public void ReadItem_NoAuto_AfterDelay_KitServer()
            // ReSharper restore InconsistentNaming
        {
            using (new StaticEnvironment())
            {
                EasyDAClient.SharedParameters.TopicParameters.SlowdownWeight = 0.0f;
                EasyDAClient.SharedParameters.TopicParameters.SpeedupWeight = 0.0f;
                var client = new EasyDAClient();
                client.InstanceParameters.HoldPeriods.TopicRead = 86400000;
                client.InstanceParameters.UpdateRates.ReadAutomatic = Timeout.Infinite;
                client.Isolated = false;
                client.InstanceParameters.Mode.AllowAsynchronousMethod = true;

                var server = new ServerDescriptor("OPCLabs.KitServer.2");
                var item = new DAItemDescriptor("Demo.Ramp");

                DAVtq vtq1 = client.ReadItem(server, item);
                AssertUtilities.WriteLine(vtq1);

                AssertUtilities.WaitingPhase(25 * 1000);

                DAVtq vtq2 = client.ReadItem(server, item);
                AssertUtilities.WriteLine(vtq2);

                AssertUtilities
                    .WaitingPhase(
                        10 * 1000); // intentional - give time to see potentially unwanted creation of new groups
            }
        }

        // Used for manually testing the internal behavior - whether OPC Groups are properly managed
        [Ignore] // IMPROVE: fails
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        // ReSharper disable InconsistentNaming
        public void ReadItem_NoAuto_AfterDelay_KepServerEx()
            // ReSharper restore InconsistentNaming
        {
            var server = new ServerDescriptor("Kepware.KEPServerEX.V5");
            LegacyTestEnvironment.Current.Precheck(server);

            using (new StaticEnvironment())
            {
                EasyDAClient.SharedParameters.TopicParameters.SlowdownWeight = 0.0f;
                EasyDAClient.SharedParameters.TopicParameters.SpeedupWeight = 0.0f;
                var client = new EasyDAClient();
                client.InstanceParameters.HoldPeriods.TopicRead = 86400000;
                client.InstanceParameters.UpdateRates.ReadAutomatic = Timeout.Infinite;
                client.Isolated = false;
                client.InstanceParameters.Mode.AllowAsynchronousMethod = true;

                var item = new DAItemDescriptor("Channel1.Device1.Tag1");

                DAVtq vtq1 = client.ReadItem(server, item);
                AssertUtilities.WriteLine(vtq1);

                AssertUtilities.WaitingPhase(25 * 1000);

                DAVtq vtq2 = client.ReadItem(server, item);
                AssertUtilities.WriteLine(vtq2);

                AssertUtilities
                    .WaitingPhase(
                        10 * 1000); // intentional - give time to see potentially unwanted creation of new groups
            }
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest()
        {
            ReadItemTest("");
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_Remote()
        {
            ReadItemTest("T-OPC");
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void ReadItemTest_WaitThusDisconnect()
        {
            ReadItemTest("");
            Thread.Sleep(60 * 1000);
            // IMPROVE: test that it is truly disconnected now
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void ReadItemTest_WaitThusDisconnect_TOPServer()
        {
            ReadItemTest("", "SWToolbox.TOPServer.V6", "Simulation.Analog.Analog_1");
            Thread.Sleep(60 * 1000);
            // IMPROVE: test that it is truly disconnected now
        }


        [Ignore] // IMPROVE: put back
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_WcfClient()
        {
            WcfClientPrecheck();
            using (new WcfClientStaticEnvironment())
                ReadItemTest("");
        }

        #endregion



        #region Private Methods

        private void BrowseNodesTest_BrowseNodesChunk(int browseNodesChunk)
        {
            using (new StaticEnvironment())
            {
                EasyDAClient.SharedParameters.ClientParameters.BrowseNodesChunk = browseNodesChunk;

                var target = new EasyDAClient();

                DANodeElementCollection actualResultCollection = target.BrowseNodes(
                    new ServerDescriptor(LegacyTestDefs.ProgIds.KitServer),
                    new DANodeDescriptor("Simulation"),
                    DABrowseParameters.Empty);

                Assert.IsNotNull(actualResultCollection);
                int count = actualResultCollection.Count;
                Assert.IsTrue(count >= 163);

                //TODO:
            }
        }

        private void InternalClientParametersTest_ExecutionContext_InprocServer()
        {
            TestSection.Arrange();
            using (new StaticEnvironment())
            {
                EasyDAClient.SharedParameters.ClientParameters.ExecutionContext = 1;
                var client = new EasyDAClient { Isolated = true };
                object value;
                TestSection.Act();
                try
                {
                    value = client.ReadItem("", LegacyTestDefs.ProgIds.KitServer,
                        LegacyTestDefs.DataAccess.ItemIds.Constant_I2);
                }
                catch (OpcException opcException)
                {
                    TestSection.Assert();
                    AssertUtilities.WriteLine(opcException);
                    Assert.AreEqual(unchecked((int)0x80040154 /*Class not registered*/), opcException.ErrorCode);
                    return;
                }
                AssertUtilities.WriteLine(value);
                TestSection.Assert();
            }
            Assert.Fail("OpcException not thrown.");
        }

        private void InternalReadItemTest_UnknownResource()
        {
            try
            {
                IEasyDAClient target = new EasyDAClient();
                target.ReadItem(LegacyTestDefs.MachineNames.UnknownResource, LegacyTestDefs.ProgIds.KitServer,
                             LegacyTestDefs.DataAccess.ItemIds.Constant_R4);
                Assert.Fail();
            }
            catch (OpcException ex)
            {
                int errorCode = ex.ErrorCode;
                AssertUtilities.WriteLine(errorCode);
                Assert.IsTrue(errorCode == -2147023174 || errorCode == -1073430527 || errorCode == -1073442813 ||
                              errorCode == (unchecked((int)0X80040154)));
                // The RPC server is unavailable OR Cannot connect client OR Machine not connected OR Class not registered
            }
        }

        private void InternalReadItemTest_UnregisteredProgID()
        {
            try
            {
                IEasyDAClient target = new EasyDAClient();
                target.ReadItem("", LegacyTestDefs.ProgIds.Unregistered, LegacyTestDefs.DataAccess.ItemIds.Constant_R4);
                Assert.Fail();
            }
            catch (OpcException ex)
            {
                Assert.AreEqual(-2147221005, ex.ErrorCode); // Invalid class string
            }
        }

        private void ReadItemTest([NotNull] string machineName)
        {
            ReadItemTest(machineName, "OPCLabs.KitServer.2", "Demo.Ramp");
        }

        private void ReadItemTest([NotNull] string machineName, [NotNull] string objectId, [NotNull] string itemId)
        {
            var serverDescriptor = new ServerDescriptor(machineName, objectId);
            TestEnvironment.Precheck(serverDescriptor);

            var client = new EasyDAClient();

            var item = new DAItemDescriptor(itemId);

            DAVtq vtq = client.ReadItem(serverDescriptor, item);
            AssertUtilities.WriteLine(vtq);
        }

        #endregion
    }
}
