﻿using System;
using System.Diagnostics;
using System.Threading;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.Engine;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpc.UA.Toolkit;
using OpcLabs.EasyOpc.UA.Toolkit.ClientServer;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA
{
    class SubscribeLoadTester
    {
        #region Private Constants

        private const int SamplingInterval = 100;

        #endregion



        #region Public Constructors

        public SubscribeLoadTester(
            [NotNull] UAEndpointDescriptor endpointDescriptor, 
            bool useMultiple,
            int nodeCount,
            int duration = 60*1000,
            int delayBetweenSubscribes = 0)
        {
            DelayBetweenSubscribes = delayBetweenSubscribes;
            Duration = duration;
            EndpointDescriptor = endpointDescriptor;
            NodeCount = nodeCount;
            UseMultiple = useMultiple;
        }

        #endregion



        #region Private Types

        private class LoadState
        {
            public int NotificationCount;

            /// <inheritdoc cref="object.ToString()"/>
            public override string ToString()
            {
                return NotificationCount.ToString();
            }
        }

        #endregion



        #region Public Properties

        public int DelayBetweenSubscribes { get; }

        public int Duration { get; }

        [NotNull]
        public UAEndpointDescriptor EndpointDescriptor { get; }

        public int NodeCount { get; }

        public bool UseMultiple { get; }

        #endregion



        #region Public Methods

        public void Test()
        {
            TestEnvironment.PrecheckEndpoint(EndpointDescriptor);

            AssertUtilities.Phase("Initializing");
            var target = new EasyUAClient();

            target.DataChangeNotification += target_DataChangeNotification;
            var loadState = new LoadState();
            var monitoredItemArgumentsArray = new EasyUAMonitoredItemArguments[NodeCount];
            for (int i = 0; i < NodeCount; i++)
                monitoredItemArgumentsArray[i] = new EasyUAMonitoredItemArguments(
                    loadState,
                    EndpointDescriptor,
                    String.Format(OpcFoundation_UASampleServer.NodeIdExpandedTexts.MemoryBuffers_UInt32Format, i),
                    new UAMonitoringParameters(SamplingInterval));
            _targetException = null;

            AssertUtilities.Phase("Subscribing");
            if (UseMultiple)
                target.SubscribeMultipleMonitoredItems(monitoredItemArgumentsArray);
            else
                for (int i = 0; i < monitoredItemArgumentsArray.Length; i++)
                {
                    AssertUtilities.WriteLine(i);
                    EasyUAMonitoredItemArguments arguments = monitoredItemArgumentsArray[i];
                    Debug.Assert(arguments != null);
                    target.SubscribeMonitoredItem(arguments);
                    Thread.Sleep(DelayBetweenSubscribes);
                }

            AssertUtilities.WaitingPhase(Duration);

            AssertUtilities.Phase("Unsubscribing");
            target.UnsubscribeAllMonitoredItems();

            AssertUtilities.Phase("Evaluating");
            AssertUtilities.WriteLine($"NotificationCount: {loadState.NotificationCount}");
            AssertUtilities.WriteLine($"TargetException: {_targetException}");
            if (_targetException != null)
                throw _targetException;
            Assert.IsTrue(loadState.NotificationCount >= 2 * NodeCount);

            AssertUtilities.Done();
        }

        #endregion



        #region Private Methods

#pragma warning disable IDE1006 // Naming Styles
        private void target_DataChangeNotification(object sender, EasyUADataChangeNotificationEventArgs eventArgs)
#pragma warning restore IDE1006 // Naming Styles
        {
            try
            {
                Assert.IsNotNull(eventArgs);
                object state = eventArgs.Arguments.State;
                Assert.IsNotNull(state);
                ((LoadState)state).NotificationCount++;
                Exception exception = eventArgs.Exception;
                // ReSharper disable once UseNullPropagation
                if (exception != null)
                {
                    if ((exception is UAClientEngineException clientEngineException) && 
                        (clientEngineException.ErrorCode == UAClientErrorCodes.PublishingHalted))
                        Assert.Inconclusive("Publishing has halted, which can also be due to high load");
                }
                Assert.IsNull(exception, $"eventArgs.Exception is not null: {exception}");
            }
            catch (Exception exception)
            {
                if (_targetException == null)
                    _targetException = exception;
            }
        }

        #endregion



        #region Private Fields

        [NotNull]
        private LegacyUATestEnvironment TestEnvironment => LegacyUATestEnvironment.Current;

        [CanBeNull]
        private Exception _targetException;

        #endregion
    }
}
